/*
 * This file is part of Skyward Hub.
 *
 * Skyward Hub is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * Skyward Hub is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * Skyward Hub. If not, see <https://www.gnu.org/licenses/>.
 *
 */

#include "SkywardHubCore.h"

#include <Core/EventHandler/EventHandler.h>
#include <Core/MessageBroker/MessageBroker.h>
#include <Core/Module/Module.h>
#include <Core/ModulesManager/ModulesManager.h>
#include <Modules/SkywardHubStrings.h>

#include <QDir>
#include <QMessageBox>

SkywardHubCore::SkywardHubCore()
{
    modulesManager      = new ModulesManager();
    moduleEventHandler  = new EventHandler();
    moduleMessageBroker = new MessageBroker();
}

SkywardHubCore::~SkywardHubCore()
{
    if (modulesManager)
    {
        delete modulesManager;
    }
    if (moduleEventHandler)
    {
        delete moduleEventHandler;
    }
    if (moduleMessageBroker)
    {
        delete moduleMessageBroker;
    }
}

void SkywardHubCore::init()
{
    if (!settings.loadFromFile(SkywardHubStrings::defaultSettingsFilePath))
    {
        // Settings file not found or loaded incorrectly
        settings.reset();
        settings.setObjectName(SkywardHubStrings::settingsObjectName);
        XmlObject defaultConfig(SkywardHubStrings::skywardHubInitFileTag);
        defaultConfig.setTextValue(
            SkywardHubStrings::defaultConfigurationFileName);
        settings.addChild(defaultConfig);
    }

    checkDefaultFolders();
    loadFirstPage();
}

void SkywardHubCore::checkDefaultFolders()
{
    /*
     * If SkywardHubConfig is missing, it create all the missing folder
     */

    if (!QDir(SkywardHubStrings::defaultConfigurationFolder).exists())
    {
        QDir().mkdir(SkywardHubStrings::defaultConfigurationFolder);
    }

    if (!QFile(SkywardHubStrings::defaultSettingsFilePath).exists())
    {
        saveSettings();
    }
}

void SkywardHubCore::saveSettings()
{
    if (settings.writeToFile(SkywardHubStrings::defaultSettingsFilePath))
    {
        QMessageBox confirmMsg;
        confirmMsg.setMinimumSize(256, 256);
        confirmMsg.setText(SkywardHubStrings::settingsSavedCorrectlyMsg);
        confirmMsg.exec();
    }
}

void SkywardHubCore::loadFirstPage()
{
    QString pageToLoadName =
        settings.getChildObjectValue(SkywardHubStrings::skywardHubInitFileTag);

    if (!pageToLoadName.isEmpty())
    {
        XmlObject pageToLoad;
        if (pageToLoad.loadFromFile(SkywardHubStrings::defaultPrefabsFolder +
                                    pageToLoadName))
        {
            modulesManager->setModules(
                modulesManager->loadModuleFromXml(pageToLoad));
            return;
        }
    }

    // If first page not found, init with empty module
    modulesManager->openNewEmptyWindow();
}

MessageBroker *SkywardHubCore::getMessageBroker()
{
    return moduleMessageBroker;
}

EventHandler *SkywardHubCore::getEventHandler() { return moduleEventHandler; }

ModulesManager *SkywardHubCore::getModulesManager() { return modulesManager; }

//_________________ SkywardHubCoreProxy
//____________________________________________________________

SkywardHubCore *SkywardHubCoreProxy::core = nullptr;

SkywardHubCore *SkywardHubCoreProxy::getCore()
{
    if (SkywardHubCoreProxy::core == nullptr)
    {
        SkywardHubCoreProxy::core = new SkywardHubCore();
    }
    return SkywardHubCoreProxy::core;
}

//_________________ ModulesManagerProxy
//____________________________________________________________

XmlObject SkywardHubCore::getSettings() const { return settings; }
