/*
 * This file is part of Skyward Hub.
 *
 * Skyward Hub is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * Skyward Hub is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * Skyward Hub. If not, see <https://www.gnu.org/licenses/>.
 *
 */

#include "CommandSelector.h"

#include <Modules/CommandPad/MessagesList.h>

#include <QDebug>

CommandSelector::CommandSelector(DefaultModule* _parent)
    : QDialog(), parent(_parent)
{
    setupUi();
}

CommandSelector::~CommandSelector()
{
    for (auto key : formElements.keys())
        delete formElements[key];
}

XmlObject CommandSelector::toXmlObject(XmlObject& obj)
{
    bool ok;
    long timeout = continuosTimeoutEdit->text().toUInt(&ok);
    if (ok)
        obj.addAttribute("timeout", QString() + timeout);

    auto key = messagesListComboBox->currentText();
    if (formElements.contains(key))
    {
        obj.addAttribute("message_id", key);
        obj.addAttribute("label", lineEdit->text());
        formElements[key]->toXmlObject(obj);
    }

    return obj;
}

void CommandSelector::fromXmlObject(const XmlObject& obj)
{
    QString str = obj.getAttribute("timeout", "0");
    long timeout;
    bool ok;
    timeout = str.toUInt(&ok);

    if (ok)
        continuosTimeoutEdit->setText(QString() + timeout);

    if (obj.hasAttribute("message_id"))
    {
        auto key = obj.getAttribute("message_id");

        if (formElements.contains(key))
        {
            formElements[key]->fromXmlObject(obj);
            messagesListComboBox->setCurrentText(key);

            selectedMessage = formElements[key]->prepareMessage(key);
            selectedLabel   = obj.getAttribute("label");
            lineEdit->setText(selectedLabel);
            selected = !selectedLabel.isEmpty();
        }
    }
}

bool CommandSelector::getSelection(QString& label, Message& message,
                                   long& continuosSendTimeout)
{
    message              = selectedMessage;
    label                = selectedLabel;
    continuosSendTimeout = continuosCheck->isChecked()
                               ? continuosTimeoutEdit->text().toUInt(nullptr)
                               : 0;
    return selected;
}

void CommandSelector::setupUi()
{
    outerLayout = new QVBoxLayout;
    outerLayout->setSizeConstraint(QLayout::SetFixedSize);

    lineEdit = new QLineEdit;
    lineEdit->setPlaceholderText("Write the button label");
    outerLayout->addWidget(lineEdit);

    continuosCheck = new QCheckBox;
    continuosCheck->setText("Continuos Send");
    outerLayout->addWidget(continuosCheck);
    QObject::connect(
        continuosCheck, &QCheckBox::stateChanged,
        [this]()
        { continuosTimeoutEdit->setHidden(!continuosCheck->isChecked()); });

    continuosTimeoutEdit = new QLineEdit();
    continuosTimeoutEdit->setPlaceholderText("Choose resend timeout in ms");
    continuosTimeoutEdit->setHidden(true);
    outerLayout->addWidget(continuosTimeoutEdit);

    messagesListComboBox = new QComboBox;
    messagesListComboBox->setSizePolicy(
        QSizePolicy(QSizePolicy::Expanding, QSizePolicy::Fixed));
    messagesListComboBox->addItems(MessagesList::messagesList);
    outerLayout->addWidget(messagesListComboBox);

    formGridLayout = new QGridLayout;
    outerLayout->addLayout(formGridLayout);

    outerLayout->addStretch();

    sendButton = new QPushButton("Select");
    outerLayout->addWidget(sendButton);

    setLayout(outerLayout);

    MessagesList::fillMessagesMap(formElements);

    // Set default value
    currentMessage = "PING_TC";
    messagesListComboBox->setCurrentText(currentMessage);

    connect(messagesListComboBox, &QComboBox::currentTextChanged, this,
            [=](QString key)
            {
                if (formElements.contains(key))
                {
                    formElements[currentMessage]->removeFromGridLayout(
                        formGridLayout);
                    formElements[key]->applyToGridLayout(formGridLayout);
                    currentMessage = key;
                }
            });

    connect(sendButton, &QPushButton::clicked,
            [=]()
            {
                if (continuosCheck->isChecked())
                {
                    bool ok;
                    continuosTimeoutEdit->text().toUInt(&ok);
                    if (!ok)
                    {
                        parent->error("Invalid Input",
                                      "Please enter a valid unsigned integer.");
                        return;
                    }
                }

                if (lineEdit->text().isEmpty())
                {
                    parent->error("Invalid Input",
                                  "Please enter a name for the button.");
                    return;
                }

                auto key = messagesListComboBox->currentText();
                if (formElements.contains(key))
                {
                    selectedMessage = formElements[key]->prepareMessage(key);
                    selectedLabel   = lineEdit->text();
                    selected        = !selectedLabel.isEmpty();

                    if (selected)
                        this->close();
                }
            });
}
