/*
 * This file is part of Skyward Hub.
 *
 * Skyward Hub is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * Skyward Hub is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * Skyward Hub. If not, see <https://www.gnu.org/licenses/>.
 *
 */

#include "OrientationVisualizer.h"

OrientationVisualizer::OrientationVisualizer(QWidget *parent)
    : DefaultModule(parent)
{

    setupUi();
    defaultContextMenuSetup();

    updateOrientation(0, 0, 0, 1);

    getCore()->getMessageBroker()->subscribe(
        Filter::fromString("Mav/ROCKET_FLIGHT_TM"), this,
        [this](const Message &message, const Filter &filter)
        {
            updateOrientation(message.getField("nas_qx").getDouble(),
                              message.getField("nas_qy").getDouble(),
                              message.getField("nas_qz").getDouble(),
                              message.getField("nas_qw").getDouble());
        });
    getCore()->getMessageBroker()->subscribe(
        Filter::fromString("Mav/PAYLOAD_FLIGHT_TM"), this,
        [this](const Message &message, const Filter &filter)
        {
            updateOrientation(message.getField("nas_qx").getDouble(),
                              message.getField("nas_qy").getDouble(),
                              message.getField("nas_qz").getDouble(),
                              message.getField("nas_qw").getDouble());
        });
}

QWidget *OrientationVisualizer::toWidget() { return this; }

XmlObject OrientationVisualizer::toXmlObject()
{
    XmlObject obj(getName(ModuleId::ORIENTATION_VISUALIZER));
    return obj;
}

void OrientationVisualizer::fromXmlObject(const XmlObject &xmlObject) {}

void OrientationVisualizer::updateOrientation(double qx, double qy, double qz,
                                              double qw)
{
    float yaw, pitch, roll;

    QQuaternion quaternion(qw, qx, qy, qz);
    quaternion.getEulerAngles(&pitch, &yaw, &roll);

    yawLabel->setText(QString::asprintf("Y: %.2lf", yaw));
    pitchLabel->setText(QString::asprintf("P: %.2lf", pitch));
    rollLabel->setText(QString::asprintf("R: %.2lf", roll));

    QQuaternion displayQuaternion = QQuaternion::fromRotationMatrix(
        quaternion.toRotationMatrix() * sceneRotation.toRotationMatrix());

    rocketTransform->setRotation(displayQuaternion);
}

void OrientationVisualizer::setupUi()
{
    Qt3DExtras::Qt3DWindow *view = new Qt3DExtras::Qt3DWindow;
    QWidget *windowContainer     = QWidget::createWindowContainer(view);

    Qt3DCore::QEntity *rootEntity = new Qt3DCore::QEntity;

    // Cone
    Qt3DCore::QEntity *rocketEntity = new Qt3DCore::QEntity(rootEntity);

    Qt3DExtras::QConeMesh *rocketMesh = new Qt3DExtras::QConeMesh();
    rocketMesh->setBottomRadius(10);
    rocketMesh->setBottomRadius(3);
    rocketMesh->setLength(20);
    rocketMesh->setRings(100);
    rocketMesh->setSlices(20);

    Qt3DExtras::QPhongMaterial *rocketMaterial =
        new Qt3DExtras::QPhongMaterial();
    rocketMaterial->setDiffuse(QColor(QRgb(0x15628c)));

    rocketTransform = new Qt3DCore::QTransform();

    rocketEntity->addComponent(rocketMesh);
    rocketEntity->addComponent(rocketMaterial);
    rocketEntity->addComponent(rocketTransform);

    // Axes lines
    drawLine({0, 0, 0}, {100, 0, 0}, QColorConstants::Red,
             rootEntity);                                                  // X
    drawLine({0, 0, 0}, {0, 100, 0}, QColorConstants::Green, rootEntity);  // Y
    drawLine({0, 0, 0}, {0, 0, 100}, QColorConstants::Blue, rootEntity);   // Z

    // Spheres

    // Qt3DExtras::QSphereMesh *sphereMesh = new Qt3DExtras::QSphereMesh;
    // sphereMesh->setRadius(3);
    // sphereMesh->setGenerateTangents(true);

    // Qt3DCore::QEntity *sphereEntity1 = new Qt3DCore::QEntity(rootEntity);
    // Qt3DCore::QTransform *sphereTransform1 = new Qt3DCore::QTransform;
    // sphereTransform1->setTranslation(QVector3D(0.0f, 0.0f, 5.0f));
    // Qt3DExtras::QPhongMaterial *sphereMaterial1 = new
    // Qt3DExtras::QPhongMaterial();
    // sphereMaterial1->setDiffuse(QColor(QRgb(0xff0000)));
    // sphereEntity1->addComponent(sphereMesh);
    // sphereEntity1->addComponent(sphereTransform1);
    // sphereEntity1->addComponent(sphereMaterial1);

    // Qt3DCore::QEntity *sphereEntity2 = new Qt3DCore::QEntity(rootEntity);
    // Qt3DCore::QTransform *sphereTransform2 = new Qt3DCore::QTransform;
    // sphereTransform2->setTranslation(QVector3D(0.0f, 5.0f, 0.0f));
    // Qt3DExtras::QPhongMaterial *sphereMaterial2 = new
    // Qt3DExtras::QPhongMaterial();
    // sphereMaterial2->setDiffuse(QColor(QRgb(0x00ff00)));
    // sphereEntity2->addComponent(sphereMesh);
    // sphereEntity2->addComponent(sphereTransform2);
    // sphereEntity2->addComponent(sphereMaterial2);

    // Qt3DCore::QEntity *sphereEntity3 = new Qt3DCore::QEntity(rootEntity);
    // Qt3DCore::QTransform *sphereTransform3 = new Qt3DCore::QTransform;
    // sphereTransform3->setTranslation(QVector3D(5.0f, 0.0f, 0.0f));
    // Qt3DExtras::QPhongMaterial *sphereMaterial3 = new
    // Qt3DExtras::QPhongMaterial();
    // sphereMaterial3->setDiffuse(QColor(QRgb(0x0000ff)));
    // sphereEntity3->addComponent(sphereMesh);
    // sphereEntity3->addComponent(sphereTransform3);
    // sphereEntity3->addComponent(sphereMaterial3);

    // Camera
    Qt3DRender::QCamera *camera = view->camera();
    camera->lens()->setPerspectiveProjection(45.0f, 16.0f / 9.0f, 0.1f,
                                             1000.0f);
    camera->setPosition(QVector3D(-30, -30, -30));
    camera->setViewCenter(QVector3D(0, 0, 0));
    camera->setUpVector(QVector3D(0, 0, -60));

    // Light
    Qt3DCore::QEntity *lightEntity = new Qt3DCore::QEntity(rootEntity);
    Qt3DRender::QPointLight *light = new Qt3DRender::QPointLight(lightEntity);
    light->setColor("white");
    light->setIntensity(1);
    lightEntity->addComponent(light);
    Qt3DCore::QTransform *lightTransform =
        new Qt3DCore::QTransform(lightEntity);
    lightTransform->setTranslation(camera->position());
    lightEntity->addComponent(lightTransform);

    // For camera controls
    // Qt3DExtras::QFirstPersonCameraController *camController =
    //     new Qt3DExtras::QFirstPersonCameraController(rootEntity);
    // camController->setLinearSpeed(50.0f);
    // camController->setLookSpeed(180.0f);
    // camController->setCamera(camera);

    view->setRootEntity(rootEntity);

    // Visualizer of roll/pitch/yaw
    QHBoxLayout *lower = new QHBoxLayout();

    yawLabel = new QLabel(this);
    yawLabel->setText("Y: 0.00");
    yawLabel->setAlignment(Qt::AlignCenter);
    pitchLabel = new QLabel(this);
    pitchLabel->setText("P: 0.00");
    pitchLabel->setAlignment(Qt::AlignCenter);
    rollLabel = new QLabel(this);
    rollLabel->setText("R: 0.00");
    rollLabel->setAlignment(Qt::AlignCenter);
    this->setStyleSheet("QLabel {font-weight: bold;}");

    lower->addWidget(yawLabel);
    lower->addWidget(pitchLabel);
    lower->addWidget(rollLabel);

    QVBoxLayout *container = new QVBoxLayout(this);
    container->addWidget(windowContainer);
    container->addLayout(lower);

    container->setStretch(0, 1);
    container->setStretch(1, 0);
}

void OrientationVisualizer::drawLine(const QVector3D &start,
                                     const QVector3D &end, const QColor &color,
                                     Qt3DCore::QEntity *rootEntity)
{
    auto *geometry = new Qt3DRender::QGeometry(rootEntity);

    // Position vertices (start and end)
    QByteArray bufferBytes;
    bufferBytes.resize(
        3 * 2 *
        sizeof(float));  // start.x, start.y, start.end + end.x, end.y, end.z
    float *positions = reinterpret_cast<float *>(bufferBytes.data());
    *positions++     = start.x();
    *positions++     = start.y();
    *positions++     = start.z();
    *positions++     = end.x();
    *positions++     = end.y();
    *positions++     = end.z();

    auto *buf = new Qt3DRender::QBuffer(geometry);
    buf->setData(bufferBytes);

    auto *positionAttribute = new Qt3DRender::QAttribute(geometry);
    positionAttribute->setName(
        Qt3DRender::QAttribute::defaultPositionAttributeName());
    positionAttribute->setVertexBaseType(Qt3DRender::QAttribute::Float);
    positionAttribute->setVertexSize(3);
    positionAttribute->setAttributeType(
        Qt3DRender::QAttribute::VertexAttribute);
    positionAttribute->setBuffer(buf);
    positionAttribute->setByteStride(3 * sizeof(float));
    positionAttribute->setCount(2);
    geometry->addAttribute(
        positionAttribute);  // We add the vertices in the geometry

    // Connectivity between vertices
    QByteArray indexBytes;
    indexBytes.resize(2 * sizeof(unsigned int));  // start to end
    unsigned int *indices = reinterpret_cast<unsigned int *>(indexBytes.data());
    *indices++            = 0;
    *indices++            = 1;

    auto *indexBuffer = new Qt3DRender::QBuffer(geometry);
    indexBuffer->setData(indexBytes);

    auto *indexAttribute = new Qt3DRender::QAttribute(geometry);
    indexAttribute->setVertexBaseType(Qt3DRender::QAttribute::UnsignedInt);
    indexAttribute->setAttributeType(Qt3DRender::QAttribute::IndexAttribute);
    indexAttribute->setBuffer(indexBuffer);
    indexAttribute->setCount(2);
    geometry->addAttribute(indexAttribute);  // We add the indices linking the
                                             // points in the geometry

    // Mesh
    auto *line = new Qt3DRender::QGeometryRenderer(rootEntity);
    line->setGeometry(geometry);
    line->setPrimitiveType(Qt3DRender::QGeometryRenderer::Lines);
    auto *material = new Qt3DExtras::QPhongMaterial(rootEntity);
    material->setAmbient(color);

    // Entity
    auto *lineEntity = new Qt3DCore::QEntity(rootEntity);
    lineEntity->addComponent(line);
    lineEntity->addComponent(material);

    // auto *transform = new Qt3DCore::QTransform();
    // transform->setRotation(sceneRotation);
    // lineEntity->addComponent(transform);
}