/*
 * This file is part of Skyward Hub.
 *
 * Skyward Hub is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * Skyward Hub is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * Skyward Hub. If not, see <https://www.gnu.org/licenses/>.
 *
 */

#include "ValuesConverterViewerModule.h"

#include <Core/MessageBroker/MessageBroker.h>

#include "Core/Message/Filter.h"
#include "ValuesViewerConfigPanel.h"
#include "ui_ValuesConverterViewerModule.h"

ValuesConverterViewerModule::ValuesConverterViewerModule(QWidget* parent)
    : DefaultModule(parent), ui(new Ui::ValuesConverterViewerModule)
{
    ui->setupUi(this);
    defaultContextMenuSetup();

    getCore()->getMessageBroker()->subscribe(
        Filter::fromString("*"), this,
        [this](const Message& message, const Filter& filter)
        { onMsgReceived(message); });
}

ValuesConverterViewerModule::~ValuesConverterViewerModule()
{
    clearLabels();
    getCore()->getMessageBroker()->unsubscribe(Filter::fromString("*"), this);
    delete ui;
}

QWidget* ValuesConverterViewerModule::toWidget() { return this; }

XmlObject ValuesConverterViewerModule::toXmlObject()
{
    XmlObject obj(getName(ModuleId::VALUESCONVERTERVIEWER));

    obj.addAttribute("Columns", QString::number(columns));

    for (int i = 0; i < rules.count(); i++)
    {
        obj.addChild(rules[i].toXmlObject());
    }
    return obj;
}

void ValuesConverterViewerModule::fromXmlObject(const XmlObject& xmlObject)
{
    if (xmlObject.getObjectName() == getName(ModuleId::VALUESCONVERTERVIEWER))
    {

        int col;
        if (xmlObject.getAttribute("Columns", col))
        {
            columns = col;
        }

        for (int i = 0; i < xmlObject.childCount(); i++)
        {
            XmlObject child = xmlObject.childAt(i);
            ValueElement el;
            if (el.fromXmlObject(child))
            {
                addRule(el);
            }
        }
        createLabels();
    }
}

void ValuesConverterViewerModule::onConfigureClicked()
{
    ValuesViewerConfigPanel* sPanel = new ValuesViewerConfigPanel();
    sPanel->setRules(rules);
    sPanel->setColumnsCount(columns);
    connect(sPanel, &ValuesViewerConfigPanel::configurationSaved, this,
            &ValuesConverterViewerModule::onConfigurationSaved);
    sPanel->show();
}

void ValuesConverterViewerModule::onConfigurationSaved(
    ValuesViewerConfigPanel* sPanel)
{
    setRules(sPanel->getRules());
    columns = sPanel->getColumnsCount();

    createLabels();
}

void ValuesConverterViewerModule::addCustomActionsToMenu()
{
    QAction* configure = new QAction("Configure");
    connect(configure, &QAction::triggered, this,
            &ValuesConverterViewerModule::onConfigureClicked);
    addActionToMenu(configure);
}

QLabel* ValuesConverterViewerModule::createView(const ValueElement& el) const
{
    QLabel* label = new QLabel();
    label->setText(el.getName());
    label->setAlignment(Qt::AlignCenter);
    return label;
}

void ValuesConverterViewerModule::clearLabels()
{
    while (labels.count() > 0)
    {
        QLabel* l = labels[0];
        labels.removeAll(l);
        delete l;
    }
}

void ValuesConverterViewerModule::createLabels()
{
    clearLabels();
    if (columns <= 0)
    {
        columns = minColNumber;
    }

    int col = 0;
    int row = 0;
    for (int i = 0; i < rules.count(); i++)
    {
        QLabel* label;

        int labelIndex = findLabelIndexByName(rules[i].getName());
        if (labelIndex >= 0 && labelIndex < labels.count())
        {
            label = labels[labelIndex];
        }
        else
        {
            label = createView(rules[i]);
            if (col == columns)
            {
                col = 0;
                row++;
            }

            if (rules[i].isDisplayInView())
            {
                ui->mainLayout_grid->addWidget(label, row, col);
                col++;
            }
        }
        labels.append(label);
    }
}

void ValuesConverterViewerModule::addRule(const ValueElement& el)
{
    rules.append(el);
}

void ValuesConverterViewerModule::setRules(const QList<ValueElement>& rList)
{
    clearRules();
    for (int i = 0; i < rList.count(); i++)
    {
        addRule(rList[i]);
    }
}

void ValuesConverterViewerModule::onMsgReceived(const Message& msg)
{
    Field value;

    for (int i = 0; i < rules.count(); i++)
    {
        // TODO
        // if (Filter::fromString(rules[i].getTopic()).matchMessage(msg, value))
        // {
        //     if (rules[i].updateCurrentValue(value.getString()))
        //     {
        //         QTime time(0, 0, 0, 0);
        //         uint64_t ms = msg.getField("timestamp").getUnsignedInteger();
        //         time        = time.addMSecs(ms);

        //         QString str = "(" + time.toString("HH:mm:ss") + ")\n";
        //         labels[i]->setText(rules[i].getName() + "\n" + str +
        //                            rules[i].getCurrentValue());
        //         if (rules[i].getColor() != "")
        //         {
        //             labels[i]->setStyleSheet("color:" + rules[i].getColor() +
        //                                      ";");
        //         }

        //         if (rules[i].isEmitActive())
        //         {
        //             Message convertedMsg(msg);
        //             auto valueField = Field(rules[i].getCurrentValue());
        //             convertedMsg.setField("value", valueField);
        //             convertedMsg.setTopic(Topic(rules[i].getOutputTopic()));
        //             getCore()->getMessageBroker()->publish(convertedMsg);
        //         }
        //     }
        // }
    }
}

void ValuesConverterViewerModule::clearRules()
{
    for (int i = 0; i < rules.count(); i++)
    {
        if (rules[i].getTopic() != "")
        {
            getCore()->getMessageBroker()->unsubscribe(
                Filter::fromString(rules[i].getTopic()), this);
        }
    }
    rules.clear();
}

int ValuesConverterViewerModule::findLabelIndexByName(const QString& name)
{
    for (int i = 0; i < labels.count(); i++)
    {
        if (labels[i]->text() == name)
        {
            return i;
        }
    }
    return -1;
}

bool ValuesConverterViewerModule::isTopicPresent(const QString& topic)
{
    for (int i = 0; i < rules.count(); i++)
    {
        if (rules[i].getTopic() == topic)
        {
            return true;
        }
    }
    return false;
}
